Task Management {#Tasks}
========================

@NAVIGATE_Tasks
@EMBEDTOC{TasksTOC}

Introduction to Task Management in ArangoDB {#TasksIntroduction}
================================================================

ArangoDB can execute user-defined JavaScript functions as periodic
tasks. This functionality can be used to implement recurring jobs
in the database.

Tasks in ArangoDB consist of a JavaScript snippet or function that is
executed each time the task is scheduled. Tasks can have optional
parameters, which are defined at task setup time. The parameters
specified at task setup time will be passed as arguments to the
task whenever it gets executed. Tasks have an execution frequency that
needs to be defined when the task is set up.

Tasks will be executed on the server they have been set up on. 
Tasks will not be shipped around in a cluster. A task will be 
executed in the context of the database it was created in. However,
when dropping a database, any tasks that were created in the context
of this database will remain active. It is therefore sensible to
first unregister all active tasks for a database before dropping the
database.

Tasks registered in ArangoDB will be executed until the server
gets shut down or restarted. After a restart of the server, any 
user-defined periodic tasks will be lost. 

Commands for Working with Tasks {#TasksCommands}
================================================

ArangoDB provides the following commands for working with tasks.
All commands can be accessed via the `tasks` module, which can be
loaded like this:

    require("org/arangodb/tasks");

Please note that the `tasks` module is available inside the ArangoDB
server only. It cannot be used from the ArangoShell or ArangoDB's web
interface. 

Register a task {#TasksRegister}
--------------------------------

To register a task, the JavaScript snippet or function needs to be
specified in addition to the execution frequency. Optionally, a task
can have an id and a name. If no id is specified, it will be auto-assigned
for a new task. The task id is also the means to access or unregister a
task later. Task names are informational only. They can be used to make
a task distinguishable from other tasks also running on the server.

The following server-side commands register a task. The command to be
executed is a JavaScript string snippet:

    var tasks = require("org/arangodb/tasks");
    tasks.register({
      id: "mytask-1",
      name: "this is a snippet task",
      period: 15,
      command: "require('console').log('hello from snippet task');"
    });
 
The above has register a task with id `mytask-1`, which will be executed
every 15 seconds on the server. The task will write a log message whenever
it is invoked.

Tasks can also be set up using a JavaScript callback function like this:

    var tasks = require("org/arangodb/tasks");
    tasks.register({
      id: "mytask-2",
      name: "this is a function task",
      period: 15,
      command: function () {
        require('console').log('hello from function task');
      }
    });

It is important to note that the callback function is late bound and 
will be executed in a different context than in the creation context. 
The callback function must therefore not access any variables defined 
outside of its own scope. The callback function can still define and
use its own variables.

To pass parameters to a task, the `params` attribute can be set when
registering a task. Note that the parameters are limited to datatypes
usable in JSON (meaning no callback functions can be passed as parameters
into a task):
 
    var tasks = require("org/arangodb/tasks");
    tasks.register({
      id: "mytask-3",
      name: "this is a parameter task",
      period: 15,
      command: function (params) {
        var greeting = params.greeting;
        var data = JSON.stringify(params.data);
        require('console').log('%s from parameter task: %s', greeting, data);
      },
      params: { greeting: "hi", data: "how are you?" }
    });

Unregister a task {#TasksUnregister}
------------------------------------

After a task has been registered, it can be unregistered using its id:
    
    var tasks = require("org/arangodb/tasks");
    tasks.unregister("mytask-1");

Note that unregistering a non-existing task will throw an exception.

List all tasks {#TasksGet}
--------------------------

To get an overview of which tasks are registered, there is the `get` 
method. If the `get` method is called without any arguments, it will
return a list of all tasks:

    var tasks = require("org/arangodb/tasks");
    tasks.get();

If called with a task id argument, it will return information about this
particular task:

    var tasks = require("org/arangodb/tasks");
    tasks.get("mytask-3");

